/*
 * skillchains-clock.js
 *
 * Copyright (c) 2004 by Daniel Potter.  All rights reserved.
 *
 * This file is part of the FFXI Skillchains Tool.
 *
 * The FFXI Skillchains Tool is free software; you can redistribute
 * it and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * The FFXI Skillchains Tool is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the FFXI Skillchains Tool; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Implements a Vana'diel clock in the browser window.
 *
 * Simple, minimally useful.
 */

// Name of Vana'diel days - note that these days are the value we give to the
// "day" attribute on our time marker to set the day icon.

const vanaDays = [ "fire", "earth", "water", "wind", "ice", "thunder", "light", "dark" ];

// Name of moon phases, used to look up the phase name and set the moon icon
const moonPhases = [ "new", "waxingcrescent", "firstquarter", "waxinggibbous", "full", "waninggibbous", "lastquarter", "waningcrescent" ];

var vanadielOffset = 0;

{
    // Calculate the offset to convert to Vana'diel time for the current time zone

    // I, unfortunately, have no idea who originally came up with the following.
    // Note that it has since been converted to GMT.
    
    // Earth time:       2002/06/23 15:00:00 GMT
    // Vana'diel time:   0898/02/01 00:00:00 (VST?)

    var vanaTime  = (898 * 360 + 30) * 24 * 60 * 60 * 1000 / 25;
    var baseDay = new Date();
    baseDay.setUTCFullYear(2002, 5, 23); // Set date to 2003-06-23
    baseDay.setUTCHours(15, 0, 0, 0);    // Set time to 15:00:00.0000

    vanadielOffset = baseDay.getTime() - vanaTime;
    
    // check to see if the clock is even active
    var useClock = false;
    var useMoon = true;
    
    var prefs = Components.classes["@mozilla.org/preferences-service;1"].
               getService(Components.interfaces.nsIPrefBranch);
    try {
	useClock = prefs.getBoolPref("skillchains.clock");
    } catch (ex) {
	// guess not
    }
    try {
	useMoon = prefs.getBoolPref("skillchains.moon");
    } catch (ex) {
	// guess so
    }

    if (useClock) {
	// add an onload listener
	addEventListener("load", useMoon ? function(event) {
	    startVanadielClock(true);
	} : function(event) {
	    startVanadielClock(false);
	}, false);
    }
}

var vanadielClockPanel = null;
var vanadielMoonIcon = null;
var vanadielDayIcon = null;
var vanadielTimeLabel = null;
var vanadielClockInterval = null;
var vanadielLastDay = -1;

function startVanadielClock(showMoon) {
    // show the moon on null
    if (showMoon == null) {
	showMoon = true;
    }
    if (vanadielClockInterval == null) {
	vanadielClockPanel = document.getElementById("skillchains-vanadiel-clock");
	vanadielClockPanel.style.display = "-moz-box";
	vanadielMoonIcon = document.getElementById("skillchains-vanadiel-moon");
	vanadielMoonIcon.style.display = showMoon ? "-moz-box" : "none";
	vanadielDayIcon = document.getElementById("skillchains-vanadiel-day");
	vanadielTimeLabel = document.getElementById("skillchains-vanadiel-time");
	// immediately update the clock
	vanadielLastDay = -1;
	updateVanadielClock();
	// calculate the timeout to the next second boundary
	var timeout = (new Date().getTime() + vanadielOffset) % 2400;
	// Starting at that timeout, update the clock every Vana'diel minute,
	// which is 60000/25 = 2400 milliseconds
	vanadielClockInterval = setTimeout('vanadielClockInterval = setInterval("updateVanadielClock()")', timeout);
    } else {
	// this may have just toggled the moon
	vanadielMoonIcon = document.getElementById("skillchains-vanadiel-moon");
	vanadielMoonIcon.style.display = showMoon ? "-moz-box" : "none";
    }
}

function stopVanadielClock() {
    if (vanadielClockInterval != null) {
	vanadielClockPanel = document.getElementById("skillchains-vanadiel-clock");
	vanadielClockPanel.style.display = "none";
	clearInterval(vanadielClockInterval);
	vanadielClockInterval = null;
    }
}

function updateVanadielClock() {
    // Vana'diel time runs 25 times faster than our time - this is the time
    // in "Vana'diel milliseconds" actually.
    var vanadielTime = (new Date().getTime() - vanadielOffset) * 25;
    // We only display the day, hour, and minute in the little clock
    var vanadielDay = Math.floor(vanadielTime / 86400000);
    // Only update day related stuff on - well, a new day
    if (vanadielDay != vanadielLastDay) {
	vanadielLastDay = vanadielDay;
	var vanadielYear = Math.floor(vanadielDay / 360);
	var vanadielMonth = Math.floor((vanadielDay % 360) / 30) + 1;
	var vanadielDayMonth = (vanadielDay % 30) + 1;
	var vanadielDayWeek = vanadielDay % 8;
	// calculate the current moon phase - unlike the current
	// day of the week, this isn't nice and happily rooted at 0
	// moonphase = 0 - full moon
	// moonphase = 42 - new moon
	var moonphase = (vanadielDay + 26) % 84;
	// translate that value into a percentage value
	var moonpercent = ((42 - moonphase) * 100) / 42;
	// negative values are "waning" and backwards
	var waxing = true;
	if (moonpercent < 0) {
	    waxing = false;
	    moonpercent = Math.abs(moonpercent);
	}
	// Now make it an interger
	moonpercent = Math.floor(moonpercent + 0.5);
	// Translate the phase into one of the eight phases.
	//
	// To see how the "phase counter" stacks up to a percent, run the
	// following Perl 1-liner:
	//
	// perl -e 'for($p=0;$p<84;$p++){print "$p = ".abs(int(((42-$p)*100/42)+0.5))."\%\n";}'
	//
	// Simple, huh?
	//
	var basicMoonPhase = 4; // full
	if (moonphase > 2 && moonphase < 16) {
	    basicMoonPhase = 5; // waning gibbous
	} else if (moonphase >= 18 && moonphase < 25) {
	    basicMoonPhase = 6; // last quarter
	} else if (moonphase >= 25 && moonphase < 39) {
	    basicMoonPhase = 7; // waning crescent
	} else if (moonphase >= 39 && moonphase < 47) {
	    basicMoonPhase = 0; // new
	} else if (moonphase >= 47 && moonphase < 60) {
	    basicMoonPhase = 1; // waxing crescent
	} else if (moonphase >= 60 && moonphase < 68) {
	    basicMoonPhase = 2; // first quarter
	} else if (moonphase >= 68 && moonphase < 82) {
	    basicMoonPhase = 3; // waxing gibbous
	}
	var date = getLocalString("day." + vanaDays[vanadielDayWeek]) + ", " + vanadielDayMonth + " / ";
	date += vanadielMonth + " / ";
	date += vanadielYear;
	dump("Calculated moon at " + moonpercent + "% " + (waxing ? "waxing" : "waning"));
	dump("\nPhase counter is " + moonphase + ", basic phase is " + basicMoonPhase + " (" + moonPhases[basicMoonPhase] + ")\n");
	vanadielMoonIcon.setAttribute("phase", moonPhases[basicMoonPhase]);
	vanadielMoonIcon.setAttribute("tooltiptext", moonpercent + "% " + getLocalString("moon." + moonPhases[basicMoonPhase]));
	vanadielClockPanel.setAttribute("tooltiptext", date);
	vanadielDayIcon.setAttribute("tooltiptext", date);
	vanadielDayIcon.setAttribute("day", vanaDays[vanadielDayWeek]);
	vanadielTimeLabel.setAttribute("tooltiptext", date);
    }
    var vanadielHour = Math.floor(vanadielTime / 3600000) % 24;
    var vanadielMinute = Math.floor(vanadielTime / 60000) % 60;
    var label = vanadielHour + ":";
    if (vanadielMinute < 10) {
	label += "0";
    }
    label += vanadielMinute;
    vanadielTimeLabel.setAttribute("value", label);
}
