/*
 * skillchains.js
 *
 * Copyright (c) 2004 by Daniel Potter.  All rights reserved.
 *
 * This file is part of the FFXI Skillchains JS Library.
 *
 * The FFXI Skillchains JS Library is free software; you can redistribute
 * it and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * The FFXI Skillchains JS Library is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the FFXI Skillchains JS Library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Basically, this defines the data model for the skillchains as well as an
 * API for using the data model.
 */

// Create an object for querying information about the current API version
var SkillchainsAPI = new Object();
// The major version number
SkillchainsAPI.majorVersion = 0;
// The minor version number
SkillchainsAPI.minorVersion = 5;
// The build version number
SkillchainsAPI.buildVersion = 6;

// Structure for a Skill Chain.  (Ultimate Renkeis have four elements.)
function Chain(name, icon, level, chainElement1, chainElement2, chainElement3, chainElement4) {
    if (name == null || level == null || chainElement1 == null)
	throw "Missing required values for new chain";
    this.name = name;
    this.icon = icon;
    this.level = level;
    this.chainElement1 = chainElement1;
    this.chainElement2 = chainElement2;
    this.chainElement3 = chainElement3;
    this.chainElement4 = chainElement4;
}

function ChainLink(chain, fromElement, toElement, prevChain) {
    if (chain == null || fromElement == null || toElement == null)
	throw "Missing values linking chains.";
    this.chain = chain;
    this.fromElement = fromElement;
    this.toElement = toElement;
    this.prevChain = prevChain;
}

function Element(name, icon) {
    this.name = name;
    this.icon = icon;
    this.chains = new Array();
    this.addChain = function(chain, element, prevChain) {
	this.chains.push(new ChainLink(chain, this, element, prevChain));
    }
    // Checks to see which chain, if any, can be made going from
    // this element to the other element.
    // otherElement - the other element
    // prevChain    - the previous chain, or null if this is the start of a chain
    this.findChain = function(otherElement, prevChain) {
	with (this) {
	    var i;
	    for (i = 0; i < chains.length; i++) {
		if (chains[i].toElement == otherElement) {
		    if (chains[i].prevChain != null) {
			if (chains[i].prevChain != prevChain) {
			    continue;
			}
		    }
		    return chains[i].chain;
		}
	    }
	}
	return null;
    }
}

var WS_NORMAL = 0;
var WS_EX = 1;
var WS_SP = 2;
var WS_AVATAR = 4;
var WS_ASTRALFLOW = 8;

function WeaponSkill(name, level, flags, element, element2, element3) {
    if (name == null || level == null)
	throw "Missing required values for Weapon Skill";
    if (flags & WS_AVATAR) {
	if (flags & (WS_AVATAR | WS_ASTRALFLOW) != flags) {
	    throw "Invalid flags " + flags + " for skill";
	}
    } else if (flags != 0 && flags != 1 && flags != 2)
	throw "Invalid flags " + flags + " for skill";
    this.name = name;
    this.flags = flags;
    this.level = level;
    this.element = element;
    this.element2 = element2;
    this.element3 = element3;
    // Find a chain (if any) *FROM* this skill *TO* the given skill
    this.findChainTo = function(toSkill, prevChain) {
	// Find a chain with the other weapon.  Our FIRST
	// element always take priority, then down the chain
	with (this) {
	    // some skills have no elements
	    // they're listed so they appear in the DB, but they can't be
	    // used to chain
	    if (element == null)
		return null;
	    var chain = element.findChain(toSkill.element, prevChain);
	    if (chain == null && toSkill.element2 != null) {
		chain = element.findChain(toSkill.element2, prevChain);
	    }
	    if (chain == null && toSkill.element3 != null) {
		chain = element.findChain(toSkill.element3, prevChain);
	    }
	    if (chain == null && element2 != null) {
		chain = element2.findChain(toSkill.element, prevChain);
		if (chain == null && toSkill.element2 != null) {
		    chain = element2.findChain(toSkill.element2, prevChain);
		}
		if (chain == null && toSkill.element3 != null) {
		    chain = element2.findChain(toSkill.element3, prevChain);
		}
	    }
	    if (chain == null && element3 != null) {
		chain = element3.findChain(toSkill.element, prevChain);
		if (chain == null && toSkill.element2 != null) {
		    chain = element3.findChain(toSkill.element2, prevChain);
		}
		if (chain == null && toSkill.element3 != null) {
		    chain = element3.findChain(toSkill.element3, prevChain);
		}
	    }
	    return chain;
	}
    }
}

//
// Represents a skillchain.
//
// Skillchains can also be further nested.  Nested chains take place
// below the current nested chain and increment the chain count for the
// current chain.
//
function SkillChain(fromSkill, toSkill, chain, starterChain) {
    this.name = chain.name;
    this.fromSkill = fromSkill;
    this.toSkill = toSkill;
    this.chain = chain;
    // The required starter chain, if any.  If there is no required first chain,
    // leave this null.
    this.starterChain = starterChain;
    // the array of child chains that can be made from this chain
    this.nextChain = new Array();
    // the previous chain if any
    this.prevChain = null;
    this.nest = function(chain) {
	this.nextChain.push(chain);
	chain.prevChain = this;
    }
    this.getStarter = function() {
	var starter = this;
	while (starter.prevChain != null) {
	    starter = starter.prevChain;
	}
	return starter;
    }
    // Compares only this chain to another chain, nested chains are
    // (currently) not taken into account
    this.equals = function(otherChain) {
	if (this.fromSkill != otherChain.fromSkill ||
	    this.toSkill != otherChain.toSkill ||
	    this.chain != otherChain.chain) {
	    return false;
	}
	return true;
    }
}

// Simple job object - not much more than a string, but...
function Job(name, abbr) {
    this.name = name;
    this.abbr = abbr;
}

function Weapon(name, ex, sp) {
    this.name = name;
    this.ex = ex;
    this.sp = sp;
    this.skills = new Array();
    this.addSkill = function(skill) {
	skill.weapon = this;
	this.skills.push(skill);
    }
    // Find a chain with another weapon
    // otherWeapon - the weapon to find a chain TO
    // maxLevel    - the maximum skill level available on either weapon
    // prevChain   - the previous chain, or null if this is the start
    this.findChains = function(otherWeapon, maxLevel, prevChain) {
	if (maxLevel == null)
	    maxLevel = 500;
	with (this) {
	    var i, j;
	    var chains = new Array();
	    for (i = 0; i < skills.length; i++) {
		if (skills[i].level <= maxLevel && skills[i].element != null) {
		    for (j = 0; j < otherWeapon.skills.length; j++) {
			if (otherWeapon.skills[j].level <= maxLevel && otherWeapon.skills[j].element != null) {
			    var chain = skills[i].findChainTo(otherWeapon.skills[j], prevChain);
			    if (chain != null) {
				chains.push(new SkillChain(skills[i], otherWeapon.skills[j], chain));
			    }
			}
		    }
		}
	    }
	    return chains;
	}
    }
}

// The player object represents a player with a given weapon at a given
// skill level
//   weapon     - the weapon being used
//   skillLevel - the maximum skill level we can use
//   ex         - true if we can use EX skills
//   sp         - true if we can use SP skills
function Player(weapon, skillLevel, ex, sp) {
    this.weapon = weapon;
    this.skillLevel = skillLevel;
    this.ex = ex;
    this.sp = sp;
    // we start out active
    this.active = true;
    // Can we use this skill? (NOTE: does *not* check to see if the skill
    // comes from the weapon used.)
    this.canUse = function(skill) {
	if (skill.level <= this.skillLevel) {
	    if (skill.flags == WS_EX) {
		return this.ex;
	    } else if (skill.flags == WS_SP) {
		return this.sp;
	    }
	    return true;
	}
	return false;
    }
    // Finds all chains possible *FROM* us *TO* the other player
    // otherPlayer - the other player
    // prevChain   - the previous chain (if any)
    // chains      - an array to append results to (optional)
    this.findChains = function(otherPlayer, prevChain, chains) {
	with (this) {
	    if (chains == null)
		chains = new Array();
	    for (var i in weapon.skills) {
		// can we use this skill?
		if (canUse(weapon.skills[i])) {
		    for (var j in otherPlayer.weapon.skills) {
			if (otherPlayer.canUse(otherPlayer.weapon.skills[j])) {
			    var chain = weapon.skills[i].findChainTo(otherPlayer.weapon.skills[j], prevChain);
			    if (chain != null) {
				chains.push(new SkillChain(weapon.skills[i], otherPlayer.weapon.skills[j], chain));
			    }
			}
		    }
		}
	    }
	    return chains;
	}
    }
    // Find chains FROM a weapon skill TO this player
    this.findChainsFromWeaponSkill = function(weaponSkill, prevChain) {
	with (this) {
	    var chains = new Array();
	    for (var i in weapon.skills) {
		// can we use this skill?
		if (canUse(weapon.skills[i])) {
		    var chain = weaponSkill.findChainTo(weapon.skills[i], prevChain);
		    if (chain != null) {
			chains.push(new SkillChain(weaponSkill, weapon.skills[i], chain));
		    }
		}
	    }
	    return chains;
	}
    }
}

// Compares two skillchains (most useful for sorting - Array.sort(skillChainCompare)).
// Skill chains are first compared by the chain number, then chain type
// (chains with one element sort earlier than chains with two, etc.), then by
// the chain type's name, then by the first skill's name, then the second skill's name.
function skillChainCompare(a, b) {
    // Sort first by chain number...
    /*if (a.chainNumber != b.chainNumber) {
	return a.chainNumber - b.chainNumber;
    }*/
    // Then by chain...
    if (a.chain != b.chain) {
	// Chains with fewer elements go first
	if (a.chain.chainElement4 == null && b.chain.chainElement4 != null)
	    return -1;
	if (a.chain.chainElement4 != null && b.chain.chainElement4 == null)
	    return 1;
	if (a.chain.chainElement2 == null && b.chain.chainElement2 != null)
	    return -1;
	if (a.chain.chainElement2 != null && b.chain.chainElement2 == null)
	    return 1;
	return a.chain.name < b.chain.name ? -1 : (a.chain.name == b.chain.name ? 0 : 1);
    } else {
	// then by weapon skill
	var aSkill = a.fromSkill, bSkill = b.fromSkill;
	if (aSkill == bSkill) {
	    aSkill = a.toSkill;
	    bSkill = b.toSkill;
	}
	return aSkill.name < bSkill.name ? -1 : (aSkill.name == bSkill.name ? 0 : 1);
    }
}

/*
 * Compares two object by their name field
 */
function compareNames(a, b) {
    return a.name < b.name ? -1 : (a.name == b.name ? 0 : 1);
}

/*
 * Takes an array of players and calculates all the chains that the players
 * can possibly create.
 *
 * NOTE: This function resets all players to the "active" state.
 */
function calculatePlayerChains(players) {
    var chains = new Array();
    var i, j, k, oldLength;
    // First make sure all the players are considered active.
    for (i in players) {
	players[i].active = true;
    }
    // Now look for chains that the players can create.
    oldLength = 0;
    //dump("Starting player chains...\n");
    for (i in players) {
	for (j in players) {
	    if (i != j) {
		//dump("Finding chains for " + players[i].weapon.name + " to " + players[j].weapon.name + "\n");
		players[i].findChains(players[j], null, chains);
		// Look for nested chains, first deactivating these players
		players[i].active = false;
		players[j].active = false;
		for (k = oldLength; k < chains.length; k++) {
		    calculateNestedChains(players, players[j], chains[k]);
		}
		oldLength = chains.length;
		// Reactivate these players
		players[j].active = true;
		players[i].active = true;
	    }
	}
    }
    return chains;
}

/*
 * Finds any available nested chains based on the players currently marked
 * as active in the players array, starting from the current player and
 * using the current chain.
 */
function calculateNestedChains(players, currentPlayer, currentChain) {
    var i, j, chains;
    for (i in players) {
	if (players[i].active) {
	    //dump("Finding chains for " + players[i].weapon.name + "\n");
	    chains = players[i].findChainsFromWeaponSkill(currentChain.toSkill, currentChain.chain);
	    // Nest these chains into the current chain.
	    for (j in chains) {
		// is this chain actually valid?  Can only go from one chain
		// to another chain of the same level or higher
		if (currentChain.chain.level <= chains[j].chain.level) {
		    currentChain.nest(chains[j]);
		}
	    }
	    // Now look for futher nested chains that can be made
	    players[i].active = false;
	    for (j in chains) {
		calculateNestedChains(players, players[i], chains[j]);
	    }
	    players[i].active = true;
	}
    }
}
